#include <iostream>
using namespace std;
#include <mpi.h>


/*	Bestimmt Min und Max (incl. Indizes) des Feldes x	*/

void MinMaxD(const int n, const double x[], 
		double& min, int& kmin, double& max, int& kmax)
{
 max = min = x[0];
 kmax = kmin = 0;
 for (int i=1; i<n; ++i )
   {
     if (x[i]>max) { max = x[i]; kmax = i;}
       else  if (x[i]<min) { min = x[i]; kmin = i;}
   }
}


/*				Exchange glob. Min. and Max. of a vector */

void PivotD(const int n, double x[])
{
/*			mm[0] - loc. minimum ,  mm[1] - loc. maximum	*/
/*			km : local indices				*/
/*			gm[0] - glob. minimum , gm[1] - glob. maximum	*/
/*			gmid : process-id with min/max			*/
 int    km[2], gmid[2];
 double mm[2], gm[2], *mg=0;
 
 MinMaxD(n,x,mm[0],km[0],mm[1],km[1]);		/* local min/max 	*/
 /*	printf("\n %g  %d   %g  %d \n",mm[0],km[0],mm[1],km[1]);	*/

 const int root=0;
 const MPI::Intracomm& icomm = MPI::COMM_WORLD;
 int numprocs = icomm.Get_size();		/*  # processes    */
 int myrank   = icomm.Get_rank();		/*  my rank number */

 if (myrank==root)  				/* require aux. array	*/
   { 
     mg = new double [2*numprocs];
   }

 icomm.Gather(mm,2,MPI::DOUBLE,mg,2,MPI::DOUBLE,root);
 
 /*	DebugD(2*numprocs,mg);	*/

 if (myrank==root) 			/* root determines glob. Min/Max */
  { 				
   gm[0] = mg[0]; gmid[0] = 0;		/*  global min 	*/
   for (int i=1; i<numprocs; ++i) 
     if ( mg[2*i] < gm[0] ) { gm[0]=mg[2*i]; gmid[0]=i; }
 				
   gm[1] = mg[1]; gmid[1] = 0;			/*  global max 	*/
   for (int i=1; i<numprocs; ++i) 
     if ( mg[2*i+1] > gm[1] ) { gm[1]=mg[2*i+1]; gmid[1]=i; }
     
      /* printf("\nroot :  %g  %d   %g  %d \n",gm[0],gmid[0],gm[1],gmid[1]); */ 
  }
  
/*			Broadcast those processes with Min or Max	*/

 icomm.Bcast(gmid,2,MPI::INT,root);
 
/*		gmid[0] and gmid[1]  exchange the Min or Max, resp.	*/

 if (gmid[0]==gmid[1] && myrank==gmid[0]) 	/* just local exchange */
  { swap( x[km[0]], x[km[1]] );  return; }

 MPI::Status stat;
 if (myrank == gmid[0]) 
  {
   icomm.Send(&x[km[0]],1,MPI::DOUBLE,gmid[1],0);
   icomm.Recv(&x[km[0]],1,MPI::DOUBLE,gmid[1],1,stat);
  }
  
 if (myrank == gmid[1])
  {
   double tt = x[km[1]];
   icomm.Recv(&x[km[1]],1,MPI::DOUBLE,gmid[0],0,stat);
   icomm.Send(&tt,1,MPI::DOUBLE,gmid[0],1);
  } 

 delete [] mg;
}
