#include <iostream>
#include <cmath>
using namespace std;
#include <mpi.h>
#include "header.h"

void JacobiSolve(const int nx, const int ny, 
	const int neigh[], const int color, const MPI::Intracomm& icomm, 
	const double sk[], const int id[], const int ik[], 
	const double f[], double u[] )
{
  const double omega   = 1.0;
  const int    maxiter = 100;
  const double tol  = 1e-5,				/* tolerance   */
               tol2 = tol*tol;				/* tolerance^2 */
  
  const int nnode = (nx+1)*(ny+1);
  double *dd = new double [nnode],			/* matrix diagonal */
         *r  = new double [nnode],			/* residual */
	 *w  = new double [nnode];			/* correction */
  
  AccuDiag(nx, ny, sk, id, dd, neigh, color, icomm); /* Acuum. diag */
//   DebugD(nnode,dd);
	 
  /*  Choose initial guess  */
  const double dzero=0.0;
  vdcopy(nnode, u, 1, &dzero, 0);		/*  u := 0         */

  /*  Initial sweep  */
  vdcopy(nnode, r, 1, f, 1);			/*                 */
  CrsMult(0, nnode-1, r, u, id, ik, sk, -1.0);	/*  r := f - K*u   */
  
  vddiv(nnode, w, 1, r, 1, dd, 1);		/*  w := D^{-1}*r  */
  VecAccu(nx,ny,w,neigh,color,icomm);		/*  dist.-->acc.   */ 
  double sigma0 = skalar(nnode,w,r,icomm);	/* s0 := <w,r>     */   
  
  /* Iteration sweeps */
  int iter  = 0;
  double sigma = sigma0;
  while ( sigma > tol2*sigma0 && maxiter>iter)
  {
    ++iter;
    vdaxpy(nnode, u, u, omega, w );		/*  u := u + om*w  */  
    vdcopy(nnode, r, 1, f, 1);			/*                 */
    CrsMult(0, nnode-1, r, u, id, ik, sk, -1.0);/*  r := f - K*u   */

    vddiv(nnode, w, 1, r, 1, dd, 1);		/*  w := D^{-1}*r  */ 
    VecAccu(nx,ny,w,neigh,color,icomm);		/*  dist.-->acc.   */ 
    sigma = skalar(nnode,w,r,icomm);		/* s0 := <w,r>     */
    
//     if (icomm.Get_rank()==0)
//     	cout << "Iteration " << iter << " : " << sqrt(sigma/sigma0) << endl;     
  }
  if (icomm.Get_rank()==0)
  	cout << "aver. Jacobi rate :  " << exp(log(sqrt(sigma/sigma0))/iter) << endl;
 
  delete [] w; delete [] r; delete [] dd;
  return;
}
