//		MPI code in C++.
//		See [Gropp/Lusk/Skjellum, "Using MPI", p.33/41 etc.]
//		and  /opt/mpich/include/mpi2c++/comm.h  for details

#include <iostream>
#include <cmath>
using namespace std;
#include <mpi.h>
#include "header.h"

int main(int argc, char* argv[])
{
//     int   nx, ny, nnx, nny, *id, *ik, nnode, sksize, ScaleUp,
//       i, zero, NXglob, NYglob,
//       ierr;
//     double *sk, dzero=0., *u, *f,
//       omega,  done = 1.;

    const bool bdebug=false;		// debugging at the end??

    MPI::Init(argc,argv);
    MPI::Intracomm icomm(MPI::COMM_WORLD);
    int numprocs = icomm.Get_size();
    int myrank   = icomm.Get_rank();

    if (myrank == 0)
    { cout << "\n There are " << numprocs << " processes running.\n \n"; }
    icomm.Barrier();

    const int procx = static_cast<int>(sqrt(numprocs+0.0));
    const int procy = procx;

    if (procy * procx != numprocs)
    {
      cout << "\n Wrong number of processors !\n \n";
    }
    else
    {
      int neigh[4], color;
      IniGeom(myrank, procx, procy, neigh, color);

      icomm.Barrier();
      /*                                              subdomain coordinates */
      double xl,xr,yb,yt;
      IniCoord(myrank, procx, procy, xl, xr, yb, yt);

      bool ScaleUp = !true;
      int nx,ny,NXglob,NYglob;
      /* number of local intervals on (xl,xr)=:nx, (yb,yt)=:ny */
      if (ScaleUp)
      {
      /*		for scale up evaluation   */
        if (myrank==0) cout << "scale up evaluation\n";
	nx = 10;
	ny = 10;
	NXglob = nx*procx;
	NYglob = ny*procy;
      }
      else
      {
      /*		for speed up evaluation   */
        if (myrank==0) cout << "speed up evaluation\n";
	NXglob = NYglob = 120;
	nx = NXglob/procx;
	ny = NYglob/procy;
      }

      /* number of nodes per direction (xl,xr)=nx+1=:nex, (yb,yt)=ny+1=:ney */
      const int nnx = nx + 1,
                nny = ny + 1,
      /* number of degrees of freedom  */
                nnode = nnx*nny;

      /* allocate memory */
      double *u = new double [nnode],          /* temperature       */
             *f = new double [nnode];          /* r.h.s.            */

      const int sksize = 4*nx*ny+2*(nx+ny)+1;	/* symm. storage     */
      double *sk = new double [sksize];		/* K                 */
      int    *ik = new int [sksize],		/* col pointer for K */
             *id = new int [nnode];		/* row pointer for K */

      /* matrix, rhs generation including Dirichlet B.C. */

      GetMatrix(nx, ny, xl, xr, yb, yt, sk, id, ik, f); /* generate K, rhs   */

      SetU(nx, ny, u);			     /* set solution/boundary values */
      ApplyDirichletBC(nx, ny, neigh, u, sk, id, ik, f); /* apply B.C.       */
//       DebugMatrix(nnode, sk, id, ik);

      JacobiSolve(nx, ny, neigh, color,icomm, sk, id, ik, f, u );

      SaveVectorP(myrank, "t.dat.", u, nx, ny, xl, xr, yb, yt);

      if (bdebug)
      {
        /*  simple check of CrsMult function	*/
     	/* check matrix, rhs */
        DebugMatrix(nnode, sk, id, ik);
        DebugD(nnode, f);

        for (int i=0; i<nnode; ++i)  u[i] = i;

	CrsMult(0, nnode-1, f, u, id, ik, sk, 0.0);
        DebugD(nnode,f);

	/*  simple check of VecAccu function	*/
	/* initialize u  */
        for (int i=0; i<nnode; ++i)  u[i] = myrank*100+i;

        DebugD(nnode,u);
        VecAccu(nx,ny,u,neigh,color,icomm);
        DebugD(nnode,u);
      }

      delete [] id; delete [] ik; delete [] sk; delete [] f; delete [] u;
  }

  MPI::Finalize();
}


