#include <stdio.h>
#include <malloc.h>
#include "mpi.h"
#include "mg.h"

void AllocLevels(nlevels, nx, ny, pp, ierr)
int 		nlevels, nx, ny, *ierr;
PointerStruct 	*pp;
{
 double dzero=0.0;
 int    mx, my, nnode, sksize, lev, zero=0;

 *ierr = (nlevels>MAXLEVEL) ? 1 : 0 ;
 if (*ierr==1) return; 

 for (lev=0; lev <MAXLEVEL; lev++)
 {
   pp->nx[lev] = pp->ny[lev] = 0;
   pp->ik[lev] = pp->id[lev] = NULL;  
   pp->sk[lev] = pp->dd[lev] = pp->f[lev]  = pp->u[lev] = pp->aux[lev] = NULL;
 }
 
 pp->nx[0] = nx;
 pp->ny[0] = ny;
 for (lev=1; lev<nlevels; lev++) 
 {
   pp->nx[lev] = 2* pp->nx[lev-1];
   pp->ny[lev] = 2* pp->ny[lev-1];
 }
 
 for (lev=0; lev<nlevels; lev++) 
 {
   mx     = pp->nx[lev];
   my     = pp->ny[lev];
   nnode  = (mx+1) * (my+1);		    /* number of degrees of freedom  */

   sksize = 4*mx*my+2*(mx+my)+1;			/* symm. storage     */
   pp->sk[lev] = (double *)malloc(sksize*sizeof(dzero));/* K                 */
   pp->ik[lev] = (int *) malloc(sksize*sizeof(zero));	/* crs pointer for K */
   pp->id[lev] = (int *) malloc(nnode*sizeof(zero));	/* crs pointer for K */
   pp->u[lev]  = (double *) malloc(nnode*sizeof(dzero));/* temperature       */
   pp->f[lev]  = (double *) malloc(nnode*sizeof(dzero));/* r.h.s.  	     */
   pp->dd[lev] = (double *) malloc(nnode*sizeof(dzero));/* diagonale         */
   pp->aux[lev]= (double *) malloc(nnode*sizeof(dzero));/* temp. vector      */   
   if (pp->sk[lev]==NULL || pp->ik[lev]==NULL || pp->id[lev]==NULL || 
       pp->u[lev] ==NULL || pp->f[lev] ==NULL || pp->dd[lev]==NULL || 
                                                pp->aux[lev]==NULL   ) *ierr=1;
 }

 return;
}


void IniLevels(xl, xr, yb, yt, neigh, color, myid, icomm, nlevels, pp, ierr)
double          xl, xr, yb, yt;
int 		*neigh,color, myid, nlevels, *ierr;
MPI_Comm icomm;
PointerStruct 	*pp;
{
 double *sk, *u, *f, *dd, dzero=0.0;
 int    *ik, *id, nx, ny, nnode, sksize, lev, zero=0;

 *ierr = 0;
 
 for (lev=0; lev<nlevels; lev++) 
 {
   nx    = pp->nx[lev];
   ny    = pp->ny[lev];
   nnode = (nx+1) * (ny+1);
   u     = pp->u[lev];

   if (lev==nlevels-1) 
     { SetU(nx, ny, u);			/* Dirichlet B.C. on fine grid */
     }
   else 
     { vdcopy(nnode,u,1,&dzero,0);	/* Hom. Dirichlet B.C. for correction */
     }
 }
 
 for (lev=0; lev<nlevels; lev++) 
 {
   nx = pp->nx[lev];
   ny = pp->ny[lev];   
   u  = pp->u[lev];
   f  = pp->f[lev];
   sk = pp->sk[lev];
   ik = pp->ik[lev];
   id = pp->id[lev];
   dd = pp->dd[lev];
      
   GetMatrix(nx, ny, xl, xr, yb, yt, sk, id, ik, f); /* generate K, rhs   */

   ApplyDirichletBC1(nx, ny, neigh, u, sk, id, ik, f); /* apply B.C.       */

   AccuDiag(nx, ny, sk, id, dd, neigh, color, myid, icomm); 

/*  
   printf("\nLevel %u :\n",lev+1);     
   DebugMatrix((nx+1)*(ny+1), sk, id, ik); 
   DebugD((nx+1)*(ny+1),f); 
*/
 } 
 return;
}


void FreeLevels(nlevels, pp, ierr)
int 		nlevels, *ierr;
PointerStruct 	*pp;
{
 int    lev;

 *ierr = 0 ;

/* for (lev=nlevels-1; lev>=0; lev--)  */
 for (lev=0; lev<=nlevels-1; lev++) 
 {
   free(pp->aux[lev]);
   free( pp->dd[lev]);
   free(  pp->f[lev]);
   free(  pp->u[lev]);
   free( pp->id[lev]);
   free( pp->ik[lev]);
   free( pp->sk[lev]);
 }
 
 return;
}
