#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include "mpi.h" 
#include "mg.h"

int main(argc,argv)
int argc;
char *argv[];
{
    MPI_Comm icomm;
    int   nx, ny, nnx, nny, *id, *ik, nnode, sksize, ScaleUp,
      i, zero, procx, procy, numprocs, myid, NXglob, NYglob,
      neigh[4], color, nlevels, ierr;
    double xl, xr, yb, yt, *sk, dzero=0., *u, *f, 
      omega,  done = 1.,
      t_start, t_mgm, usertime();
    ControlStruct control;
    PointerStruct pointers;
    
 
    icomm = MPI_COMM_WORLD;
    MPI_Init(&argc,&argv);
    MPI_Comm_size(icomm,&numprocs);
    MPI_Comm_rank(icomm,&myid);
    if (myid == 0)
    {
       printf("\n There are %d processes running.\n \n",numprocs);
    }
    
    MPI_Barrier(icomm);

    procx = (int) sqrt((double) numprocs);
    procy = (int) sqrt((double) numprocs);

    if (procy * procx != numprocs) 
    {
      printf("\n Wrong number of processors !\n \n");
    }
    else
    {
      IniGeom(myid, procx, procy, neigh, &color);

      MPI_Barrier(icomm);
      /*                                              subdomain coordinates */
      IniCoord(myid, procx, procy, &xl, &xr, &yb, &yt);

      nlevels = 5; 				/* set number of mg levels */
      
      ScaleUp = -1;
      /* number of local intervals on (xl,xr)=:nx, (yb,yt)=:ny */
      if (ScaleUp==1)
      {
      /*		for scale up evaluation   */ 
        if (myid==0) printf("scale up evaluation\n");
	nx = 2;
	ny = 2;
	NXglob = nx*procx;
	NYglob = ny*procy;
      }
      else
      {
      /*		for speed up evaluation   */ 
        if (myid==0) printf("speed up evaluation\n");
	NXglob = NYglob = 8; 
	nx = NXglob/procx;
	ny = NYglob/procy;
	if ( NXglob % procx != 0   || NYglob % procy != 0 )
	   { MPI_Finalize(); 
	     printf("\n Check discretization vs. number of processors\n");
	     exit(-1); 
	   }
 	nlevels -= (int)(log((double)procx)/log(2.0)); 
	if (myid==0)
	   printf("%u level on %u x %u processors. \n",nlevels,procx,procy);
      }
      
      			t_start = usertime();
      
      /* Set mg control parameters */
      control.cycle = 1;		/* V-cycle			*/
      control.npre  = 1;		/* #sweeps in presmoothing	*/
      control.npost = 1;		/* #sweeps in postsmoothing	*/
            
      AllocLevels(nlevels, nx, ny, &pointers, &ierr);
      if (ierr!=0) printf("\nError No. %i in AllocLevels.\n",ierr);
      
      IniLevels(xl, xr, yb, yt, neigh, color, myid, icomm, 
      		nlevels, &pointers, &ierr);

      MGMSolver(&pointers, &control, nlevels-1, neigh,color,myid,icomm);

      MPI_Barrier(icomm);
      			if (myid==0) printf("-----------\n"); 
      FreeVecAccu();
      FreeLevels(nlevels, &pointers, &ierr); 
    
      			t_mgm = usertime()-t_start;   
  }
  MPI_Barrier(icomm);
  			if (myid==0) printf("End of program: %g sec.\n",t_mgm);
  MPI_Finalize();
}

            
