#include <stdio.h>
#include <malloc.h>
#include <stdlib.h>
#include "mpi.h"
#include "mg.h"

void AllocLevels(nlevels, nx, ny, pp, ierr)
int 		nlevels, nx, ny, *ierr;
PointerStruct 	*pp;
{
 double dzero=0.0;
 int    mx, my, nnode, sksize, lev, zero=0;

 *ierr = (nlevels>MAXLEVEL) ? 1 : 0 ;
 if (*ierr==1) return; 

 for (lev=0; lev <MAXLEVEL; lev++)
 {
   pp->nx[lev] = pp->ny[lev] = 0;
   pp->ik[lev] = pp->id[lev] = NULL;  
   pp->sk[lev] = pp->dd[lev] = pp->f[lev]  = pp->u[lev] = pp->aux[lev] = NULL;
 }
 
 pp->nx[0] = nx;
 pp->ny[0] = ny;
 for (lev=1; lev<nlevels; lev++) 
 {
   pp->nx[lev] = 2* pp->nx[lev-1];
   pp->ny[lev] = 2* pp->ny[lev-1];
 }
 
 for (lev=0; lev<nlevels; lev++) 
 {
   mx     = pp->nx[lev];
   my     = pp->ny[lev];
   nnode  = (mx+1) * (my+1);		    /* number of degrees of freedom  */
//	printf("%u  %u\n",my,nnode);

   sksize = 4*mx*my+2*(mx+my)+1;			/* symm. storage     */
   pp->sk[lev] = (double *)malloc(sksize*sizeof(dzero));/* K                 */
   pp->ik[lev] = (int *) malloc(sksize*sizeof(zero));	/* crs pointer for K */
   pp->id[lev] = (int *) malloc(nnode*sizeof(zero));	/* crs pointer for K */
   pp->u[lev]  = (double *) malloc(nnode*sizeof(dzero));/* temperature       */
   pp->f[lev]  = (double *) malloc(nnode*sizeof(dzero));/* r.h.s.  	     */
   pp->dd[lev] = (double *) malloc(nnode*sizeof(dzero));/* diagonale         */
   pp->aux[lev]= (double *) malloc(nnode*sizeof(dzero));/* temp. vector      */   
   if (pp->sk[lev]==NULL || pp->ik[lev]==NULL || pp->id[lev]==NULL || 
       pp->u[lev] ==NULL || pp->f[lev] ==NULL || pp->dd[lev]==NULL || 
                                                pp->aux[lev]==NULL   ) *ierr=1;
 }
 return;
}


void IniLevels(xl, xr, yb, yt, neigh, color, myid, icomm, nlevels, pp, ierr)
double          xl, xr, yb, yt;
int 		*neigh,color, myid, nlevels, *ierr;
MPI_Comm icomm;
PointerStruct 	*pp;
{
 double *sk, *u, *f, *dd, dzero=0.0;
 int    *ik, *id, nx, ny, nnode, sksize, lev, zero=0;

 *ierr = 0;
 
 for (lev=0; lev<nlevels; lev++) 
 {
   nx    = pp->nx[lev];
   ny    = pp->ny[lev];
   nnode = (nx+1) * (ny+1);
   u     = pp->u[lev];

   if (lev==nlevels-1) 
     { SetU(nx, ny, u);			/* Dirichlet B.C. on fine grid */
     }
   else 
     { vdcopy(nnode,u,1,&dzero,0);	/* Hom. Dirichlet B.C. for correction */
     }
 }
 
 for (lev=0; lev<nlevels; lev++) 
 {
   nx = pp->nx[lev];
   ny = pp->ny[lev];   
   u  = pp->u[lev];
   f  = pp->f[lev];
   sk = pp->sk[lev];
   ik = pp->ik[lev];
   id = pp->id[lev];
   dd = pp->dd[lev];
      
   GetMatrix(nx, ny, xl, xr, yb, yt, sk, id, ik, f); /* generate K, rhs   */

   ApplyDirichletBC1(nx, ny, neigh, u, sk, id, ik, f); /* apply B.C.       */

   AccuDiag(nx, ny, sk, id, dd, neigh, color, myid, icomm); 

/*  
   printf("\nLevel %u :\n",lev+1);     
   DebugMatrix((nx+1)*(ny+1), sk, id, ik); 
   DebugD((nx+1)*(ny+1),f); 
*/
 } 
 return;
}


void FreeLevels(nlevels, pp, ierr)
int 		nlevels, *ierr;
PointerStruct 	*pp;
{
 int    lev;

 *ierr = 0 ;

 
 free(pp->csk);
 free(pp->cik);
 free(pp->cid);
 free(pp->loc2glob);
 
 
/* for (lev=nlevels-1; lev>=0; lev--)  */
 for (lev=0; lev<=nlevels-1; lev++) 
 {
   free(pp->aux[lev]);
   free( pp->dd[lev]);
   free(  pp->f[lev]);
   free(  pp->u[lev]);
   free( pp->id[lev]);
   free( pp->ik[lev]);
   free( pp->sk[lev]);
 }
 
 return;
}

int Comp(ii,jj)
const void *ii,*jj;
{
 int k=0, *i, *j;
 i = (int*)ii;  j = (int*)jj;
 if (*i<*j) k = -1;
 else if (*i>*j) k = 1;
 return k;
}


void MergePatterns(numprocs, nsk, id1, id2, hh, cik, kk, myid)
int  numprocs, nsk, id1, id2, myid, 
     hh[], cik[], kk[];
{
 int i,k,nk,n,tk;
 int Comp();
 
 n  = id2 - id1+1;
 nk = 0; 

 for (i=0; i<numprocs; i++)
   if (hh[i*nsk+id1]>=0) { vicopy(n,&kk[nk],1,&hh[i*nsk+id1],1);  nk+=n; }

 qsort(kk, nk, sizeof(int), Comp);
 
/* if (myid==0)
 { printf(": ");
   for(i=0;i<nk;i++) printf(" %i  ",kk[i]);
   printf("\n");
 } */
 
 k  = id1;
 tk = -1;
 for (i=0;i<nk;i++)
 {
  if (kk[i]>tk)
  { 
    tk = kk[i];
    cik[k] = tk; 
    k++;
  }
 }

 for (i=k; i<=id2; i++)  cik[i] = -1;

/* if (myid==0)
 { printf("# ");
   for(i=id1;i<=id2;i++) printf(" %i  ",cik[i]);
   printf("\n");
 } */
  
 return;
}


void GlobalCoarse(neigh, color, myid, icomm, nlevels, pp, ierr)
int 		*neigh,color, myid, nlevels, *ierr;
MPI_Comm        icomm;
PointerStruct 	*pp;
{
 double *r, *w, *csk, *sk, *tt, dzero=0.0, done=1.0, skalar(), t;
 int    nx,ny,nnode,bitid,i,j,k,id1,id2,ic,jc,imax, nnew, numprocs, myoffset, 
        nglob, offs, izero=0, bi, bj, nsk, cid1,cid2,cij,ij, maxentries, 
        *mark, *gnnew, *l2g, *cid, *cik, *id, *ik, *hh, *kk; 

 MPI_Comm_size(icomm,&numprocs);
 nx   = pp->nx[0];
 ny   = pp->ny[0];
 nnode= (nx+1)*(ny+1);
 r    = pp->aux[0]; 
 w    = pp->u[0];
 sk   = pp->sk[0];
 ik   = pp->ik[0];
 id   = pp->id[0];
 mark = (int*)malloc(nnode*sizeof(int));
 gnnew= (int*)malloc(numprocs*sizeof(int));
 
 l2g = pp->loc2glob = (int*)malloc(nnode*sizeof(int));

/*				Determine global number of nodes	*/
 vdcopy(nnode,r,1,&done,0); 
 VecAccu(nx,ny,r,neigh,color,myid,icomm);
 vddiv(nnode,r,1,&done,0,r,1);			/* r := diag(R^{-1}) */
 
 vdcopy(nnode,w,1,&done,0);
 pp->nglob = (int)skalar(nnode,w,r,icomm);    /* nglob: glob. # coarse points */
 nglob = pp->nglob; 

/* printf("local #points: %u    global #points: %u\n",nnode,nglob ); */

/* following code only correct until 32 processors  (int is 4 Byte long) !!  */
/* count number of new coarse points: nnew		*/

 bitid = 1 << myid; 				/* set bit-id  2^myid */
			/* printf("proc %u with bit-id  %u\n",myid,i); */
 t = (double)bitid;
 vdcopy(nnode,r,1,&t,0);
 VecAccu(nx,ny,r,neigh,color,myid,icomm);	/* r: acc. bit-id */
						/* DebugD(nnode,r); */
 for (i=0; i<nnode; i++)
    mark[i] = ( ((int)r[i]) % bitid == 0 ) ? 1 : 0; 
 nnew = 0;
 for (i=0; i<nnode; i++)  nnew += mark[i];

/* Calculate offset for global numbering */
 MPI_Allgather(&nnew, 1, MPI_INT, gnnew, 1, MPI_INT, icomm);
 
 myoffset=0;
 for (i=0;i<myid; i++)  myoffset += gnnew[i];
 
 				/* printf("%u --> %u\n", myid, myoffset); */

/* local to global numbering fixed*/ 

 offs = myoffset;
 for (i=0; i<nnode; i++)
     if (mark[i]==1) l2g[i] = i+offs;
     else          { l2g[i] = 0; offs--; }

 for (i=0; i<nnode; i++)  r[i] = (double)l2g[i];
 VecAccu(nx,ny,r,neigh,color,myid,icomm);
 for (i=0; i<nnode; i++)  l2g[i] = (int)r[i]; 
							/* DebugI(nnode,l2g); */
/*  determine global matrix pattern */ 
 cid = pp->cid = (int*)malloc(nglob*sizeof(int));
 
 hh = (int*)malloc(nglob*sizeof(int));			/* aux. vector */
 vicopy(nglob,hh,1,&dzero,0);
							/*  CRSs matrix */
 id2 = -1; 
 for (i=0; i<nnode; i++) 
 {
    ic   = l2g[i];
    id1  = id2 + 1;
    id2  = id[i];
    for (j = id1+1; j <=id2; j++) 			/* without diagonal */
    {
       {  jc = l2g[ik[j]];
	  if (jc>=ic)  hh[ic]++; 
	  if (jc<ic)   hh[jc]++;
       }
    }
 }
 MPI_Allreduce(hh, cid, nglob, MPI_INT, MPI_SUM, icomm);
						/*  DebugI(nglob,hh);  */
						/*  DebugI(nglob,cid); */
 for (i=1; i<nglob; i++)  cid[i] += (cid[i-1]+1);
 nsk = cid[nglob-1]+1;
/*  Now, cid[i] points to end of row i */
						/*  printf("%i\n",nsk); */
						/*  DebugI(nglob,cid);  */
 maxentries = cid[0];
 for (i=2; i<nglob; i++)  
 	if (cid[i]-cid[i-1]>maxentries)  maxentries=cid[i]-cid[i-1];
 						
 cik = pp->cik  = (int*)malloc(nsk*sizeof(int)); 
 for (i=0; i<nsk; i++)  cik[i] = -1; 

 hh  = (int*)realloc(hh,nglob*sizeof(int)); 		/*  aux. vector  */
 hh[0] = 0;
 for (i=1; i<nglob; i++)  hh[i] = cid[i-1]+1;

 id2 = -1; 						
 for (i=0; i<nnode; i++) 
 {
    ic   = l2g[i];
    id1  = id2 + 1;
    id2  = id[i];
    for (ij = id1; ij <=id2; ij++) 
    {
       j  = ik[ij];
       jc = l2g[j]; 
       if (jc>=ic)  
       		{ cik[hh[ic]] = jc; hh[ic]++; }
       if (jc<ic)   
       		{ cik[hh[jc]] = ic; hh[jc]++; }
    }
 }						
							/* DebugI(nsk,cik); */
 hh  = (int*)realloc(hh,nsk*numprocs*sizeof(int)); 	/*  aux. vector  */
 MPI_Allgather(cik, nsk, MPI_INT, hh, nsk, MPI_INT, icomm); 
 						/* DebugI(nsk*numprocs,hh); */
 kk = (int*)malloc(maxentries*numprocs*sizeof(int));	/*  aux. vector  */
 
 id2 = -1;
 for (i=0; i<nglob; i++)
 {
   id1 = id2+1;
   id2 = cid[i]; 
   MergePatterns(numprocs, nsk, id1, id2, hh, cik, kk, myid);  
 }
/* printf("Proc. %i hats geschafft: %i  \n",myid,maxentries*numprocs); */

/*  compress cik, e.g., delete entries=-1	*/

 cid2 = -1;
 for (i=0; i<nglob; i++)
 {
  cid1 = cid2+1;
  cid2 = cid[i];
  offs = 0;
  for (j=cid1; j<=cid2; j++)  if (cik[j]<0) offs++;
  if (offs>0)
  {
    cij = cid2-offs;
    vicopy(nsk-cid2, &cik[cij+1], 1, &cik[cid2+1], 1);
    nsk -= offs;
    cid2 = cij;
    for (j=i; j<nglob; j++)  cid[j] -= offs;
  }
 }
	/* if (myid==0) printf("new nsk: %i  %i \n",nsk, cid[nglob-1]); */
							/* DebugI(nsk,cik);  */
 free(kk); 
 free(hh); 
 free(gnnew); 
 free(mark); 
 
 cik = pp->cik  = (int*)realloc(pp->cik,nsk*sizeof(int)); 

 csk = pp->csk = (double*)malloc(nsk*sizeof(double));
 tt  = (double*)malloc(nsk*sizeof(double));		/* aux. vector */

 vdcopy(nsk,tt,1,&dzero,0);
 id2 = -1;
 for (i=0; i<nnode; i++)
 {
  ic   = l2g[i];
  id1  = id2+1;
  id2  = id[i];
  cid1 = (ic==0) ? 0 : cid[ic-1]+1;
  cid2 = cid[ic];
  for (ij=id1; ij<=id2; ij++)
  {
   j   = ik[ij];
   jc  = l2g[j];
   cij = -1;
   for (k=cid1; k<=cid2; k++)
      if (jc==cik[k])  cij = k;
   if (k<0)  printf("\n something's WRONG in GlobalCoarse !!\n");
   tt[cij] = sk[ij];
  }
 }

 MPI_Allreduce(tt, csk, nsk, MPI_DOUBLE, MPI_SUM, icomm);

			/* if (myid==0) DebugMatrix(nglob,csk,cid,cik); */

 free(tt);
 
 return;
}
