/*
	void GetMatrix(nx, ny, xl, xr, yb, yt, sk, id, ik)

	Computes matrix for Poisson eqution. 
	a) Determine coordinates and element connectivity matrix wrt. 
		an equidistant discretization.
	b) We use a general code to calculate element matrices and 
		local stiffness matrix K. 

	IN:	nx, ny	points in x/y direction
		dx, dy  subdomain-size in x/y-direction
	OUT:	sk	symmetric matrix
		id, ik	crs pointer of sk (symmetric matrix)
*/
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
/* 
	numbering: rowise
*/
void GetCoordsInRectangle(nx, ny, xl, xr, yb, yt, xc)
int nx, ny;
double  xl,xr,yb,yt, xc[][2]; 
{
  double y0, hx, hy;
  int    i, j, k;
  
  hx = (xr-xl)/nx;  hy = (yt-yb)/ny;
  
  k = 0;
  for (j=0; j<=ny; j++)
  {
    y0 = yb + j*hy;
    for (i=0; i<=nx; i++, k++)
    {
      xc[k][0] = xl + i*hx;
      xc[k][1] = y0;
    }
  }

  return;
}

/* 
	element numbering: rowise - lower/upper
	node numbering in element: lex. forward
*/
void GetConnectivityInRectangle(nx, ny, ia)
int nx, ny, ia[][3];
{
  int i,j,l,k,dx;
  
  dx = nx+1;
  k  = 0;
  l  = 0;
  for (j=0; j<ny; j++, k++)
  {
    for (i=0; i<nx; i++, k++)
    {
      ia[l][0] = k;
      ia[l][1] = k+1;
      ia[l][2] = k+dx+1;
      l++;
      ia[l][0] = k;
      ia[l][1] = k+dx;
      ia[l][2] = k+dx+1;
      l++;     
    } 
  }
  return;
}

/* 
	symmetric matrix stored in CRS, end of row pointer
*/
void GetMatrixpointersInRectangle(nx, ny, id, ik)
int nx, ny, id[], ik[];
{
  int i,j,k,ij,dx; 

  dx = nx+1;
  k  = 0;					/*  node counter */
  ij = -1;					/* entry counter */
  for (j=0; j<ny; j++)
  {
    for (i=0; i<nx; i++, k++)
    {
      ij++;
      ik[ij] = k;
      ij++;
      ik[ij] = k+1;
      ij++;
      ik[ij] = k+dx;
      ij++;
      ik[ij] = k+dx+1;
      id[k]  = ij;
    }
     
      ij++;					/* right edge of domain */
      ik[ij] = k;
      ij++;
      ik[ij] = k+dx;
      id[k]  = ij; 
      k++; 
  }
  						/*   top edge of domain */
  for (i=0; i<nx; i++, k++)
  {
    ij++;
    ik[ij] = k;
    ij++;
    ik[ij] = k+1;
    id[k]  = ij;
  } 
  					/*   top right corner of domain */
    ij++;
    ik[ij] = k;
    id[k]  = ij;

  return;
}

/* 
			general routine for lin. triangular elements
*/
void CalcElem(ial, xc, ske, fe)
int    ial[3];
double xc[][2], ske[3][3], fe[3];
{
  extern double FunctF(double x,double y);

  double x13,x21,x32,y13,y21,y32,jac,xm,ym,fm;
  int i1,i2,i3;
  
  i1  = ial[0];   i2= ial[1];   i3 = ial[2];
  x13 = xc[i3][0] - xc[i1][0];  y13 = xc[i3][1] - xc[i1][1];
  x21 = xc[i1][0] - xc[i2][0];  y21 = xc[i1][1] - xc[i2][1];
  x32 = xc[i2][0] - xc[i3][0];  y32 = xc[i2][1] - xc[i3][1];
  
  jac = fabs(x21*y13-x13*y21); 
  
  ske[0][0] = 0.5/jac* (y32*y32+x32*x32);
  ske[0][1] = 0.5/jac* (y13*y32+x13*x32);
  ske[0][2] = 0.5/jac* (y21*y32+x21*x32);
  ske[1][0] = ske[0][1];
  ske[1][1] = 0.5/jac* (y13*y13+x13*x13);
  ske[1][2] = 0.5/jac* (y21*y13+x21*x13);
  ske[2][0] = ske[0][2];
  ske[2][1] = ske[1][2];
  ske[2][2] = 0.5/jac* (y21*y21+x21*x21);
  
  xm    = (xc[i1][0] + xc[i2][0] + xc[i3][0])/3;
  ym    = (xc[i1][1] + xc[i2][1] + xc[i3][1])/3;
  fm    = 0.5*jac* FunctF(xm,ym)/3.0;
  fe[0] = fe[1] = fe[2] = fm;
}

/* 
	general routine for lin. triangular elements,
	symm. matrix
	node numbering in element: lex. forward
*/
void AddElem(ial, ske, fe, sk, id, ik, f)
int    ial[3], id[], ik[];
double ske[3][3], fe[3], sk[], f[];
{
  int i,j,ip,id1,id2,ii,jj,btest;

/* printf("[%u,%u,%u] .\n",ial[0],ial[1],ial[2]);  */
  for (i=0; i<3; i++)
  {
    ii  = ial[i];
    id1 = (ii==0) ? -1 : id[ii-1];
    id2 = id[ii];
    ip  = id1;
    for (j=i; j<3; j++)
    {
      jj    = ial[j];
      btest = -1;
      do
      { ip++;
        btest = (ik[ip]==jj);
      }
      while (!btest && ip<id2);

      if (!btest) 
      	{ printf("Error in AddElem: (%u,%u) [%u,%u,%u] .\n",
					ii,jj,ial[0],ial[1],ial[2]); 
	  exit(-123);
	}
      
      sk[ip] += ske[i][j];
    }
    f[ii] += fe[i];
  }   
}

/* 
*/
void DebugMatrix(nnode, sk, id, ik)
int    nnode, id[], ik[];
double sk[];
{
  int i,j,ip,id1,id2,ii,jj,btest;

  printf("\nMatrix\n");
  id2 = -1;
  for (i=0; i<nnode; i++)
  {
    printf("Row %u : ",i);
    id1 = id2+1;
    id2 = id[i];
    for (j=id1; j<=id2; j++)
      printf("[%u]  %f , ",ik[j],sk[j]);
    printf("\n");
  }

  return;
}
/*
	void GetMatrix(nx, ny, xl, xr, yb, yt, sk, id, ik)

	Computes matrix for Poisson eqution. 
	a) Determine coordinates and element connectivity matrix wrt. 
		an equidistant discretization.
	b) We use a general code to calculate element matrices and 
		local stiffness matrix K. 

	IN:	nx, ny	points in x/y direction
		dx, dy  subdomain-size in x/y-direction
	OUT:	sk	symmetric matrix
		id, ik	crs pointer of sk
*/

void GetMatrix(nx, ny, xl, xr, yb, yt, sk, id, ik, f)
int nx, ny, *id, *ik;
double  xl,xr,yb,yt, *sk, *f;
{
  double *xc, ske[3][3], fe[3], dzero=0.0;
  int    *ia, nnode, nelem, i;
  
  nnode = (nx+1)*(ny+1);
  nelem = 2*nx*ny; 
  xc = (double *) malloc(2*nnode*sizeof(double));  /* coordinates	  */
  ia = (   int *) malloc(3*nelem*sizeof(   int));  /* connectivity matrix */
  
  GetCoordsInRectangle(nx, ny, xl, xr, yb, yt, xc);

  GetConnectivityInRectangle(nx, ny, ia);

  GetMatrixpointersInRectangle(nx, ny, id, ik);
  
  vdcopy(id[nnode-1], sk, 1, &dzero, 0);
  vdcopy(nnode, f, 1, &dzero, 0);

						/* Loop over all elements */
  for (i=0; i<nelem; i++)
  {
    CalcElem(&ia[3*i], xc, ske, fe);
    
    AddElem(&ia[3*i], ske, fe, sk, id, ik, f);
  }

/*  DebugMatrix(nnode, sk, id, ik); */

  free(ia); free(xc);
  return;
}
