#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <mpi.h>

void JacobiSolve(nx, ny, neigh, color, myid, icomm, 
		     sk, id, ik, f, u )
int    nx,ny,*neigh,color,myid,*id,*ik;
MPI_Comm icomm;
double *sk, *f, *u;
{
  extern double skalar();
  double *dd, *r, *w, dzero=0, sigma, sigma0, tol, tol2, omega;
  int nnode, iter, maxiter;
  
  omega   = 1.0;
  maxiter = 100;
  tol  = 1e-5;						/* tolerance   */
  tol2 = tol*tol;					/* tolerance^2 */
  
  nnode = (nx+1)*(ny+1); 
  dd    = (double *) malloc(nnode*sizeof(double));  /* matrix diagonal */
  r     = (double *) malloc(nnode*sizeof(double));  /* residual */
  w     = (double *) malloc(nnode*sizeof(double));  /* correction */

  AccuDiag(nx, ny, sk, id, dd, neigh, color, myid, icomm); /* Acuum. diag */
	 
  /*  Choose initial guess  */
  vdcopy(nnode, u, 1, &dzero, 0);		/*  u := 0         */

  /*  Initial sweep  */
  vdcopy(nnode, r, 1, f, 1);			/*                 */
  CrsMult(0, nnode-1, r, u, id, ik, sk, -1.0);	/*  r := f - K*u   */
  
  vddiv(nnode, w, 1, r, 1, dd, 1);		/*  w := D^{-1}*r  */
  VecAccu(nx,ny,w,neigh,color,myid,icomm);	/*  dist.-->acc.   */ 
  sigma0 = skalar(nnode,w,r,icomm);		/* s0 := <w,r>     */   
  
  /* Iteration sweeps */
  iter  = 0;
  sigma = sigma0;
  while ( sigma > tol2*sigma0 && maxiter>iter)
  {
    iter++;
    vdaxpy(nnode, u, u, omega, w );		/*  u := u + om*w  */  
    vdcopy(nnode, r, 1, f, 1);			/*                 */
    CrsMult(0, nnode-1, r, u, id, ik, sk, -1.0);/*  r := f - K*u   */

    vddiv(nnode, w, 1, r, 1, dd, 1);		/*  w := D^{-1}*r  */ 
    VecAccu(nx,ny,w,neigh,color,myid,icomm);	/*  dist.-->acc.   */ 
    sigma = skalar(nnode,w,r,icomm);		/* s0 := <w,r>     */
    
/*    if (myid==0)
    	printf("Iteration %u : %g\n",iter,sqrt(sigma/sigma0));    */ 
  }
  if (myid==0)
  	printf("aver. Jacobi rate :  %g\n", exp(log(sqrt(sigma/sigma0))/iter));
 
  free(w); free(r); free(dd);
  return;
}
