#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <mpi.h>

/*  initial guess u and accumulated diagonal are INPUT */

void JacobiSolve2(nx, ny, neigh, color, myid, icomm, 
		     sk, id, ik, f, u, dd, aux )
int    nx,ny,*neigh,color,myid,*id,*ik;
MPI_Comm icomm;
double *sk, *f, *u, *dd, *aux;
{
  extern double skalar();
  double *r, *w, dzero=0, sigma, sigma0, tol, tol2, omega;
  int nnode, iter, maxiter;
 
  omega   = 1.0;
  maxiter = 100;
  tol  = 1e-5;						/* tolerance   */
  tol2 = tol*tol;					/* tolerance^2 */
  
  nnode = (nx+1)*(ny+1);
  r = aux;
  w = (double *) malloc(nnode*sizeof(double));

  /*  Initial sweep  */
  vdcopy(nnode, r, 1, f, 1);			/*                 */
  CrsMult(0, nnode-1, r, u, id, ik, sk, -1.0);	/*  r := f - K*u   */
  
  vddiv(nnode, w, 1, r, 1, dd, 1);		/*  w := D^{-1}*r  */
  VecAccu(nx,ny,w,neigh,color,myid,icomm);	/*  dist.-->acc.   */ 
  sigma0 = skalar(nnode,w,r,icomm);		/* s0 := <w,r>     */   
  
  /* Iteration sweeps */
  iter  = 0;
  sigma = sigma0;
  while ( sigma > tol2*sigma0 && maxiter>iter)
  {
    iter++;
    vdaxpy(nnode, u, u, omega, w );		/*  u := u + om*w  */  
    vdcopy(nnode, r, 1, f, 1);			/*                 */
    CrsMult(0, nnode-1, r, u, id, ik, sk, -1.0);/*  r := f - K*u   */

    vddiv(nnode, w, 1, r, 1, dd, 1);		/*  w := D^{-1}*r  */ 
    VecAccu(nx,ny,w,neigh,color,myid,icomm);	/*  dist.-->acc.   */ 
    sigma = skalar(nnode,w,r,icomm);		/* s0 := <w,r>     */
    
/*    if (myid==0)
    	printf("Iteration %u : %g\n",iter,sqrt(sigma/sigma0));   */ 
  } 
  MPI_Barrier(icomm);
/*  printf("Process %u : %u Iterations\n",myid,iter); */
  
  free(w);
  return;
}
