#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <mpi.h>

void JacobiSolveGlob(myid, icomm, nnode, nglob, csk, cid, cik, l2g, f, u )
int      myid,nnode,nglob,*cid,*cik,*l2g;
MPI_Comm icomm;
double   *csk, *f, *u;
{
  extern double dscapr();
  double *dd, *cr, *cf, *cu, 
         dzero=0, sigma, sigma0, tol, tol2, omega;
  int    iter, maxiter, i, ia;
  
  omega   = 1.0;
  maxiter = 100;
  tol  = 1e-5;						/* tolerance   */
  tol2 = tol*tol;					/* tolerance^2 */
    
  cf    = (double *) malloc(nglob*sizeof(double));	/* r.h.s.   */
  cr    = (double *) calloc(nglob,sizeof(double));	/* residual */

  for (i=0; i<nnode; i++)  cr[l2g[i]] = f[i];
  
  MPI_Allreduce(cr, cf, nglob, MPI_DOUBLE, MPI_SUM, icomm);
  
  dd    = (double *) malloc(nglob*sizeof(double));	/* matrix diagonal */
  ia = 0;
  for (i = 0; i < nglob; i++) 
    { dd[i] = csk[ia];  ia = cid[i]+1; }
	 
  /*  Choose initial guess = 0 */
  cu    = (double *) calloc(nglob,sizeof(double));	/*  u := 0  */
  
  /*  Initial sweep  */
  vdcopy(nglob, cr, 1, cf, 1);				/*                 */
  CrsMult(0, nglob-1, cr, cu, cid, cik, csk, -1.0);	/*  r := f - K*u   */
  
  vddiv(nglob, cr, 1, cr, 1, dd, 1);		/*  r := D^{-1}*r  */
  sigma0 = dscapr(nglob,cr,cr);			/* s0 := <r,r>     */   
  
  /* Iteration sweeps */
  iter  = 0;
  sigma = sigma0;
  while ( sigma > tol2*sigma0 && maxiter>iter)
  {
    iter++;
    vdaxpy(nglob, cu, cu, omega, cr );		/*  u := u + om*r  */  
    vdcopy(nglob, cr, 1, cf, 1);		/*                 */
    CrsMult(0, nglob-1, cr, cu, cid, cik, csk, -1.0);/*  r := f - K*u   */

    vddiv(nglob, cr, 1, cr, 1, dd, 1);		/*  r := D^{-1}*r  */ 
    sigma = dscapr(nglob,cr,cr);		/* s0 := <r,r>     */
    
/*    if (myid==0)
    	printf("Iteration %u : %g\n",iter,sqrt(sigma/sigma0));    */ 
  }
/*
  if (myid==0)
  	printf("aver. Jacobi rate :  %g\n", exp(log(sqrt(sigma/sigma0))/iter));
*/ 
  for (i=0; i<nnode; i++)  u[i] = cu[l2g[i]];

  free(cu); free(dd); free(cr); free(cf);
  return;
}
